/******************************************************************************
**                                                                           **
**    PMM 0.1 - The physical memory manager                                  **
**  -----------------------------------------                                **
**                                                                           **
**    Source: pmm.c / pmm.h                                                  **
**                                                                           **
**    Purpose:                                                               **
**      This is where the initialization, allocation and deallocation        **
**      of pages happens.                                                    **
**                                                                           **
**    Author: BugHunter                                                      **
**                                                                           **
**    Contact:                                                               **
**      bughunter2@googlemail.com                                            **
**                                                                           **
**    Date: 29th of June, 2007                                               **
**                                                                           **
**    Current version: 0.1                                                   **
**                                                                           **
**    Changelog:                                                             **
**      0.1 First version                                                    **
**      0.2 Bugfix: The stack is assigned a virtual address                  **
**                  instead of a physical one.                               **
**                                                                           **
******************************************************************************/

#include "../../../libc/include/stdint.h"
#include "../../../libc/include/stddef.h"

#include "../../multiboot_header.h"
#include "../../kernelfunctions/include/text.h"
#include "pmm.h"

    /* stack and its stack top index */
    uintptr_t *pmm_page_stack;
    unsigned long pmm_page_stack_top = __def__pmm_page_stack_size;

    /* symbols defined in the linker script */
    extern uintptr_t kernel_base;
    extern uintptr_t kernel_end;
    extern uintptr_t kernel_start;

    // --- void *pmm_allocpage()
    // *
    // *  purpose:
    // *    Allocates a page.
    // *
    // *  parameters:
    // *    No parameters.
    // *
    // *  returns:
    // *    Returns a physical address to the allocated page.
    // *
    // ---
    void *pmm_allocpage() {
        /* check if there is any page free */
        if(__def__pmm_page_stack_size == pmm_page_stack_top) return NULL;

        /* return a pointer to the physical page address */
        return (void*)pmm_page_stack[pmm_page_stack_top++];
    }

    // --- void pmm_freepage()
    // *
    // *  purpose:
    // *    Frees a page.
    // *
    // *  parameters:
    // *    void *ptr = A page which was previously allocated using
    //*                 pmm_allocpage() is deallocated.
    // *
    // *  returns:
    // *    No returns.
    // *
    // ---
    void pmm_freepage(void *ptr) {
        pmm_page_stack[--pmm_page_stack_top] = (uintptr_t)ptr;
    }

    // --- int pmm_init()
    // *
    // *  purpose:
    // *    Finds memory to store the page stack at, and then calls
    // *    pmm_setup_stack() to setup the page stack.
    // *
    // *  parameters:
    // *    multiboot_info_t *mbi = A pointer to the Multiboot information
    // *                            structure passed by the Multiboot-compliant
    // *                            boot loader.
    // *
    // *  returns:
    // *    On success, a value of 1 is returned.
    // *    On failure, a value of 0 is returned.
    // *
    // ---
    int pmm_init(multiboot_info_t *mbi) {
        #ifdef DEBUG
            kputs("Setting up the physical memory manager...");
        #endif

        /* search for contiguous physical memory to manage the page stack */
        pmm_page_stack = NULL;
        memory_map_t *mmap_current = (memory_map_t*)mbi->mmap_addr;
        while(mmap_current < (memory_map_t*)(mbi->mmap_addr+mbi->mmap_length))
        {
            /* check if it's available memory and if there is enough contiguous memory in this block */
            if(1 == mmap_current->type && mmap_current->length_low >= __def__pmm_page_stack_size) {
                uintptr_t page_addr = mmap_current->base_addr_low;
                unsigned long amount = mmap_current->length_low;
                while(amount >= __def__pmm_page_size) {
                    /* <TODO> Remove the 4MiB mark, we're here selecting an address to store the
                       free page stack. We're now making sure the page stack resides in memory higher
                       than address 0x00400000 because below that address we have our kernel.
                       Put a symbol at the start of the kernel (in the linker script) and at the end,
                       then substract them and use this symbol to measure the size of the kernel.
                       So the 0x00400000 below should be something like 'kernel_start_addr+kernel_size' */
                    if(page_addr < (uintptr_t)((char*)&kernel_start-(char*)&kernel_base) || page_addr >= (uintptr_t)((char*)&kernel_end-(char*)&kernel_base)) {
                        /* we've found a place to store the page stack */
                        pmm_page_stack = (uintptr_t*)(page_addr+(char*)&kernel_base);

                        #ifdef DEBUG
                            kprintf("pmm_page_stack is located at: 0x%x\n", pmm_page_stack);
                            kprintf("pmm_page_stack ends at: 0x%x\n", pmm_page_stack+__def__pmm_page_stack_size);
                        #endif
                        break;
                    }

                    /* decrease the amount of memory that's left to check and increase the page address */
                    amount    -= __def__pmm_page_size;
                    page_addr += __def__pmm_page_size;
                }

                /* check if we've found a place for the stack */
                if(NULL != pmm_page_stack)
                    break;
            }

            /* update the pointer to the next entry */
            mmap_current = (void*)((uintptr_t)(mmap_current)+mmap_current->size+sizeof(mmap_current->size));
        }

        /* check if we've found space or not */
        if(NULL == pmm_page_stack) {
            /* no space could be found */
            kputs("ERROR: No space could be found to store the stack of free pages");
            return 0;
        }

        /* now initialize the page stack */
        if(!pmm_init_stack(mbi)) {
            kputs("ERROR: Failed setting up the physical memory manager.");
            return 0;
        }

        return 1;
    }

    // --- int pmm_init_stack()
    // *
    // *  purpose:
    // *    Initializes the page stack. This function should be called first
    // *    to initialize the page stack for first use.
    // *
    // *  parameters:
    // *    multiboot_info_t *mbi = A pointer to the Multiboot information
    // *                            structure passed by the Multiboot-compliant
    // *                            boot loader.
    // *
    // *  returns:
    // *    On success, a value of 1 is returned.
    // *    On failure, a value of 0 is returned.
    // *
    // ---
    int pmm_init_stack(multiboot_info_t *mbi) {
        /* push all available addresses onto the page stack */
        memory_map_t *mmap_current = (memory_map_t*)mbi->mmap_addr;
        while(mmap_current < (memory_map_t*)(mbi->mmap_addr+mbi->mmap_length))
        {
            /* check if it's available memory */
            if(1 == mmap_current->type) {
                /* add 'length'/4096 pages to the page stack from the current map of available memory */
                uintptr_t page_addr = mmap_current->base_addr_low;
                unsigned long amount = mmap_current->length_low;
                while(amount >= __def__pmm_page_size) {
                    /* Check if this is a page above the 16MB DMA memory mark.
                       IMPORTANT: We also make sure we do not add the pages
                       that are owned by the stack or kernel to the stack.
                       <TODO> Remove the 4MiB mark, we're here selecting an address to store the
                       free page stack. We're now making sure the page stack resides in memory higher
                       than address 0x00400000 because below that address we have our kernel.
                       Put a symbol at the start of the kernel (in the linker script) and at the end,
                       then substract them and use this symbol to measure the size of the kernel.
                       So the 0x00400000 below should be something like 'kernel_start_addr+kernel_size' */
                    if(__def__pmm_16mb_dma_mark <= page_addr &&
                       (page_addr < (uintptr_t)((char*)&kernel_start-(char*)&kernel_base) || page_addr >= (uintptr_t)((char*)&kernel_end-(char*)&kernel_base)) && /* check if page is outside the kernel's memory */
                       (page_addr < (uintptr_t)pmm_page_stack || page_addr >= (uintptr_t)(pmm_page_stack+__def__pmm_page_stack_size)) /* and check if the page does not belong to the stack itself */) {
                        /* add the current page to the page stack */
                        pmm_page_stack[--pmm_page_stack_top] = page_addr;
                    }

                    /* decrease the amount of memory that's left to check and increase the page address */
                    amount    -= __def__pmm_page_size;
                    page_addr += __def__pmm_page_size;
                }
            }

            /* update the pointer to the next entry */
            mmap_current = (void*)((uintptr_t)(mmap_current)+mmap_current->size+sizeof(mmap_current->size));
        }

        /* check if there was at least one page pushed onto the stack */
        if(pmm_page_stack_top == __def__pmm_page_stack_size) {
            /* there is no available memory, stack pointer still at the absolute top */
            kputs("ERROR: No available memory to initialize the page stack.");
            return 0;
        }

        #ifdef DEBUG
            kprintf("pmm_page_stack_top: 0x%x\n", pmm_page_stack_top);
            kprintf("amount of pages on the stack: 0x%x\n", __def__pmm_page_stack_size-pmm_page_stack_top);
        #endif

        return 1;
    }
