/******************************************************************************
**                                                                           **
**    PMM 0.1 - The physical memory manager                                  **
**  -----------------------------------------                                **
**                                                                           **
**    Source: pmm.c / pmm.h                                                  **
**                                                                           **
**    Purpose:                                                               **
**      This is where the initialization, allocation and deallocation        **
**      of pages happens.                                                    **
**                                                                           **
**    Author: BugHunter                                                      **
**                                                                           **
**    Contact:                                                               **
**      bughunter2@googlemail.com                                            **
**                                                                           **
**    Date: 29th of June, 2007                                               **
**                                                                           **
**    Current version: 0.1                                                   **
**                                                                           **
**    Changelog:                                                             **
**      0.1 First version                                                    **
**      0.2 Bugfix: The stack is assigned a virtual address                  **
**                  instead of a physical one.                               **
**                                                                           **
******************************************************************************/

#include "../../../libc/include/stdint.h"
#include "../../multiboot_header.h"

    /* This one must be aligned on __def__pmm_page_size boundary.
       Why? Because it's used to modify the 'base_addr_low' member
       of the memory_map_t structure passed by the Multiboot compliant
       boot loader. The 'base_addr_low' member is later on used to
       setup paging, so it should be aligned on __def__pmm_page_size. */
    #define __def__pmm_page_stack_size 0x00100000
    #define __def__pmm_page_size       0x00001000
    #define __def__pmm_16mb_dma_mark   0x01000000

    // --- void *pmm_allocpage()
    // *
    // *  purpose:
    // *    Allocates a page.
    // *
    // *  parameters:
    // *    No parameters.
    // *
    // *  returns:
    // *    Returns a physical address to the allocated page.
    // *
    // ---
    void *pmm_allocpage();

    // --- void pmm_freepage()
    // *
    // *  purpose:
    // *    Frees a page.
    // *
    // *  parameters:
    // *    void *ptr = A page which was previously allocated using
    //*                 pmm_allocpage() is deallocated.
    // *
    // *  returns:
    // *    No returns.
    // *
    // ---
    void pmm_freepage(void *ptr);

    // --- int pmm_init()
    // *
    // *  purpose:
    // *    Finds memory to store the page stack at, and then calls
    // *    pmm_setup_stack() to setup the page stack.
    // *
    // *  parameters:
    // *    multiboot_info_t *mbi = A pointer to the Multiboot information
    // *                            structure passed by the Multiboot-compliant
    // *                            boot loader.
    // *
    // *  returns:
    // *    On success, a value of 1 is returned.
    // *    On failure, a value of 0 is returned.
    // *
    // ---
    int pmm_init(multiboot_info_t *mbi);

    // --- int pmm_init_stack()
    // *
    // *  purpose:
    // *    Initializes the page stack. This function should be called first
    // *    to initialize the page stack for first use.
    // *
    // *  parameters:
    // *    multiboot_info_t *mbi = A pointer to the Multiboot information
    // *                            structure passed by the Multiboot-compliant
    // *                            boot loader.
    // *
    // *  returns:
    // *    On success, a value of 1 is returned.
    // *    On failure, a value of 0 is returned.
    // *
    // ---
    int pmm_init_stack(multiboot_info_t *mbi);
